from dataclasses import dataclass
from typing import Optional

from tdm.abstract.datamodel import AbstractDomainType


@dataclass(frozen=True)
class NERCRegexp(object):
    regexp: str
    context_regexp: Optional[str] = None
    auto_create: bool = False


@dataclass(frozen=True)
class NERCBasedType(AbstractDomainType):
    regexp: tuple[NERCRegexp, ...] = tuple()
    black_regexp: tuple[NERCRegexp, ...] = tuple()
    pretrained_nerc_models: tuple[str, ...] = tuple()
    dictionary: tuple[str, ...] = tuple()
    black_list: tuple[str, ...] = tuple()

    def __post_init__(self):
        if self.regexp:
            object.__setattr__(self, 'regexp', tuple(map(self._convert_to_regexp, self.regexp)))
        if self.black_regexp:
            object.__setattr__(self, 'black_regexp', tuple(map(self._convert_to_regexp, self.black_regexp)))

        for attr in ['regexp', 'black_regexp', 'pretrained_nerc_models', 'dictionary', 'black_list']:
            if isinstance(getattr(self, attr), list):
                object.__setattr__(self, attr, tuple(getattr(self, attr)))

    @staticmethod
    def _convert_to_regexp(obj) -> NERCRegexp:
        if isinstance(obj, NERCRegexp):
            return obj
        if isinstance(obj, dict):
            return NERCRegexp(**obj)
        raise ValueError



from dataclasses import dataclass

from tdm.abstract.json_schema import generate_model
from tdm.datamodel.domain import ComponentValueType as BaseComponentValueType

from ._relext import RelExtBasedType
from ._value import AtomValueType

__deps = (AtomValueType,)


@generate_model(label='component')
@dataclass(frozen=True)
class ComponentValueType(RelExtBasedType, BaseComponentValueType):
    isRequired: bool = False  # TODO: move to tdm?



from dataclasses import dataclass

from tdm.abstract.datamodel import FactStatus
from tdm.abstract.json_schema import generate_model
from tdm.datamodel.domain import RelationType as BaseRelationType
from tdm.datamodel.facts import ConceptFact, RelationFact

from ._concept import ConceptType
from ._relext import RelExtBasedType

__deps = (ConceptType,)


@generate_model(label='relation')
@dataclass(frozen=True)
class RelationType(RelExtBasedType, BaseRelationType):
    def __post_init__(self):
        BaseRelationType.__post_init__(self)

    def build_fact(self, status: FactStatus, source: ConceptFact, target: ConceptFact) -> RelationFact:
        return RelationFact(status, self, source, target)



from dataclasses import dataclass

from tdm.abstract.datamodel import FactStatus
from tdm.abstract.json_schema import generate_model
from tdm.datamodel.domain import PropertyType as BasePropertyType, RelationPropertyType as BaseRelationPropertyType
from tdm.datamodel.facts import ConceptFact, PropertyFact, ValueFact

from ._concept import ConceptType
from ._nerc import NERCBasedType
from ._relation import RelationType
from ._relext import RelExtBasedType
from ._value import AtomValueType

__deps = (AtomValueType, ConceptType, RelationType)


@generate_model(label='property')
@dataclass(frozen=True)
class PropertyType(RelExtBasedType, BasePropertyType):
    isIdentifying: bool = False

    def __post_init__(self):
        BasePropertyType.__post_init__(self)

    def build_fact(self, status: FactStatus, source: ConceptFact, target: ValueFact) -> PropertyFact:
        return PropertyFact(status, self, source, target)


@generate_model(label='id_property')
@dataclass(frozen=True)
class IdentifyingPropertyType(NERCBasedType, PropertyType):
    isIdentifying: bool = True

    def __post_init__(self):
        if not self.isIdentifying:
            raise ValueError
        PropertyType.__post_init__(self)


@generate_model(label='r_property')
@dataclass(frozen=True)
class RelationPropertyType(RelExtBasedType, BaseRelationPropertyType):
    def __post_init__(self):
        BaseRelationPropertyType.__post_init__(self)



from functools import singledispatch

from tdm.abstract.datamodel import AbstractValue
from tdm.datamodel.values import DateTimeValue, DoubleValue, GeoPointValue, IntValue, StringValue, TimestampValue


@singledispatch
def get_filters(value: AbstractValue) -> dict:
    raise NotImplementedError


@get_filters.register
def _string_filter(value: StringValue) -> dict:
    return {"stringFilter": {"str": value.value.replace('"', '?')}}  # see TKL-987


def _int_timestamp_filter(value: IntValue | TimestampValue) -> dict:
    return {"intFilter": {"start": value.value, "end": value.value}}


@get_filters.register
def _int_filter(value: IntValue) -> dict:
    return _int_timestamp_filter(value)


@get_filters.register
def _timestamp_filter(value: TimestampValue) -> dict:
    return _int_timestamp_filter(value)


@get_filters.register
def _double_filter(value: DoubleValue) -> dict:
    return {"doubleFilter": {"start": value.value, "end": value.value}}


@get_filters.register
def _date_time_filter(value: DateTimeValue) -> dict:
    date_filter = {"date": value.date.__dict__}
    if value.time:
        date_filter["time"] = value.time.__dict__

    return {"dateTimeFilter": {"start": date_filter, "end": date_filter}}


@get_filters.register
def _geo_filter(value: GeoPointValue) -> dict:
    geo_filter = {"radius": 0.0001}
    if value.point:
        geo_filter["point"] = value.point.__dict__
    if value.name:
        geo_filter["name"] = value.name.__dict__

    return {"dateTimeFilter": {"start": geo_filter, "end": geo_filter}}



from dataclasses import dataclass
from typing import Optional

from tdm.abstract.json_schema import generate_model
from tdm.datamodel.domain import AtomValueType as BaseAtomValueType, CompositeValueType

from ._nerc import NERCBasedType


@generate_model(label='atom')
@dataclass(frozen=True)
class AtomValueType(NERCBasedType, BaseAtomValueType):
    value_restriction: Optional[tuple[str, ...]] = None

    def __post_init__(self):
        BaseAtomValueType.__post_init__(self)
        if isinstance(self.value_restriction, list):
            object.__setattr__(self, 'value_restriction', tuple(self.value_restriction))


generate_model(label='composite')(CompositeValueType)



from dataclasses import dataclass
from typing import Optional

from tdm.abstract.datamodel import AbstractDomainType


@dataclass(frozen=True)
class RelExtModel(object):
    relation_type: str
    invert_direction: bool = False
    source_annotation: Optional[str] = None
    target_annotation: Optional[str] = None


@dataclass(frozen=True)
class RelExtBasedType(AbstractDomainType):
    pretrained_relext_models: tuple[RelExtModel, ...] = tuple()

    def __post_init__(self):
        if self.pretrained_relext_models:
            object.__setattr__(self, 'pretrained_relext_models', tuple(map(self._convert_to_model, self.pretrained_relext_models)))
        elif isinstance(self.pretrained_relext_models, list):
            object.__setattr__(self, 'pretrained_relext_models', tuple(self.pretrained_relext_models))

    @staticmethod
    def _convert_to_model(obj) -> RelExtModel:
        if isinstance(obj, RelExtModel):
            return obj
        if isinstance(obj, dict):
            return RelExtModel(**obj)
        raise ValueError
